@@ -2,19 +2,41 @@
 MeshedRegion
 ============
 """
-from ansys import dpf
-from ansys.dpf.core import scoping
-from ansys.dpf.core.check_version import server_meet_version
-from ansys.dpf.core.common import locations, types
+import traceback
+import warnings
+
+import ansys.dpf.core.errors
+
+from ansys.dpf.core import scoping, field, property_field
+from ansys.dpf.core.check_version import server_meet_version, version_requires
+from ansys.dpf.core.common import locations, types, nodal_properties
 from ansys.dpf.core.elements import Elements, element_types
 from ansys.dpf.core.nodes import Nodes
-from ansys.dpf.core.plotter import Plotter as _DpfPlotter
+from ansys.dpf.core.plotter import DpfPlotter, Plotter
 from ansys.dpf.core.cache import class_handling_cache
-from ansys.grpc.dpf import meshed_region_pb2, meshed_region_pb2_grpc
+from ansys.dpf.core import server as server_module
+from ansys.dpf.gate import meshed_region_capi, meshed_region_grpcapi
+
+
+def update_grid(func):
+    # Decorate mesh setters to centralize the update logic of pyvista objects.
+    def wrapper(*args, **kwargs):
+        mesh = args[0]
+        if mesh._full_grid is not None:
+            # Treat each setter separately to improve performance by updating the minimum required.
+            if func.__name__ == 'set_coordinates_field':
+                # When setting node coordinates
+                from ansys.dpf.core.vtk_helper import vtk_update_coordinates
+                vtk_update_coordinates(vtk_grid=mesh._full_grid, coordinates_array=args[1].data)
+
+        return func(*args, **kwargs)
+    return wrapper
+
 
 @class_handling_cache
 class MeshedRegion:
-    """Represents a mesh from DPF.
+    """
+    Represents a mesh from DPF.
 
     Parameters
     ----------
@@ -29,21 +51,13 @@ class MeshedRegion:
         The default is ``None``, in which case an attempt is made to use the
         global server.
 
-    Attributes
-    ----------
-    nodes : Nodes
-        Entity containing all nodal properties.
-
-    elements : Elements
-        Entity containing all elemental properties.
-
     Examples
     --------
     Extract a meshed region from a model.
 
     >>> import ansys.dpf.core as dpf
     >>> from ansys.dpf.core import examples
-    >>> model = dpf.Model(examples.static_rst)
+    >>> model = dpf.Model(examples.find_static_rst())
     >>> meshed_region = model.metadata.meshed_region
 
     Create a meshed region from scratch (line with 3 beam elements).
@@ -66,22 +80,30 @@ class MeshedRegion:
     """
 
     def __init__(self, num_nodes=None, num_elements=None, mesh=None, server=None):
+        # step 1: get server
+        self._server = server_module.get_or_create_server(server)
 
-        if server is None:
-            server = dpf.core._global_server()
+        # step 2: get api
+        self._api = self._server.get_api_for_type(
+            capi=meshed_region_capi.MeshedRegionCAPI,
+            grpcapi=meshed_region_grpcapi.MeshedRegionGRPCAPI,
+        )
 
-        self._server = server
-        self._stub = self._connect()
+        # step3: init environment
+        self._api.init_meshed_region_environment(self)  # creates stub when gRPC
 
-        if isinstance(mesh, MeshedRegion):
-            self._message = mesh._mesh
-        elif isinstance(mesh, meshed_region_pb2.MeshedRegion):
-            self._message = mesh
-        elif mesh is None:
-            self.__send_init_request(num_nodes, num_elements)
-        else:  # support_pb2.Support
-            self._message = meshed_region_pb2.MeshedRegion()
-            self._message.id = mesh.id
+        # step4: if object exists: take instance, else create it:
+        # object_name -> protobuf.message, DPFObject*
+        if mesh is not None:
+            self._internal_obj = mesh
+        else:
+            # if no mesh object, create one
+            if self._server.has_client():
+                self._internal_obj = self._api.meshed_region_new_on_client(
+                    self._server.client
+                )
+            else:
+                self._internal_obj = self._api.meshed_region_new()
 
         self._full_grid = None
         self._elements = None
@@ -99,14 +121,31 @@ class MeshedRegion:
         scoping : Scoping
             ids of the elements or nodes of the mesh
         """
-        request = meshed_region_pb2.GetScopingRequest(mesh=self._message)
-        request.loc.location = loc
-        out = self._stub.GetScoping(request)
-        return scoping.Scoping(scoping=out, server=self._server)
+        if loc is locations.nodal:
+            out = self._api.meshed_region_get_shared_nodes_scoping(self)
+        elif loc is locations.elemental:
+            out = self._api.meshed_region_get_shared_elements_scoping(self)
+        else:
+            raise TypeError(f"Location {loc} is not recognized.")
+        if out is None:
+            return
+        scop_to_return = scoping.Scoping(scoping=out, server=self._server)
+        try:
+            check = scop_to_return._api.scoping_fast_access_ptr(scop_to_return)
+            if check is None:
+                return None
+        except NotImplementedError:
+            # will throw NotImplementedError for ansys-grpc-dpf
+            pass
+        except ansys.dpf.core.errors.DPFServerException:
+            # and DPFServerException for gRPC CLayer
+            pass
+        return scop_to_return
 
     @property
     def elements(self):
-        """All elemental properties of the mesh, such as connectivity and element types.
+        """
+        All elemental properties of the mesh, such as connectivity and element types.
 
         Returns
         -------
@@ -117,20 +156,19 @@ class MeshedRegion:
         --------
         >>> import ansys.dpf.core as dpf
         >>> from ansys.dpf.core import examples
-        >>> model = dpf.Model(examples.static_rst)
+        >>> model = dpf.Model(examples.find_static_rst())
         >>> meshed_region = model.metadata.meshed_region
         >>> elements = meshed_region.elements
         >>> print(elements)
         DPF Elements object with 8 elements
 
         """
-        if self._elements is None:
-            self._elements = Elements(self)
-        return self._elements
+        return Elements(self)
 
     @property
     def nodes(self):
-        """All nodal properties of the mesh, such as node coordinates and nodal connectivity.
+        """
+        All nodal properties of the mesh, such as node coordinates and nodal connectivity.
 
         Returns
         -------
@@ -141,20 +179,19 @@ class MeshedRegion:
         --------
         >>> import ansys.dpf.core as dpf
         >>> from ansys.dpf.core import examples
-        >>> model = dpf.Model(examples.static_rst)
+        >>> model = dpf.Model(examples.find_static_rst())
         >>> meshed_region = model.metadata.meshed_region
         >>> nodes = meshed_region.nodes
         >>> nodes.n_nodes
         81
 
         """
-        if self._nodes is None:
-            self._nodes = Nodes(self)
-        return self._nodes
+        return Nodes(self)
 
     @property
     def unit(self):
-        """Unit of the meshed region.
+        """
+        Unit of the meshed region.
 
         This unit is the same as the unit of the coordinates of the meshed region.
 
@@ -166,7 +203,8 @@ class MeshedRegion:
 
     @unit.setter
     def unit(self, value):
-        """Unit type.
+        """
+        Unit type.
 
         Parameters
         ----------
@@ -175,44 +213,99 @@ class MeshedRegion:
         return self._set_unit(value)
 
     def _get_unit(self):
-        """Retrieve the unit type.
+        """
+        Retrieve the unit type.
 
         Returns
         -------
         unit : str
         """
-        return self._stub.List(self._message).unit
+        return self._api.meshed_region_get_unit(self)
 
     def _set_unit(self, unit):
-        """Set the unit of the meshed region.
+        """
+        Set the unit of the meshed region.
 
         Parameters
         ----------
         unit: str
         """
-        request = meshed_region_pb2.UpdateMeshedRegionRequest()
-        request.meshed_region.CopyFrom(self._message)
-        request.unit = unit
-        return self._stub.UpdateRequest(request)
+        return self._api.meshed_region_set_unit(self, unit)
 
     def __del__(self):
         try:
-            self._stub.Delete(self._message)
+            self._deleter_func[0](self._deleter_func[1](self))
         except:
-            pass
-
-    def _connect(self):
-        """Connect to the gRPC service containing the reader."""
-        return meshed_region_pb2_grpc.MeshedRegionServiceStub(self._server.channel)
+            warnings.warn(traceback.format_exc())
 
     def __str__(self):
         from ansys.dpf.core.core import _description
 
-        return _description(self._message, self._server)
+        return _description(self._internal_obj, self._server)
+
+    @property
+    def available_property_fields(self):
+        """
+        Returns a list of available property fields
+
+        Returns
+        -------
+        available_property_fields : list str
+        """
+        available_property_fields = []
+        n_property_field = self._api.meshed_region_get_num_available_property_field(
+            self
+        )
+        for index in range(n_property_field):
+            available_property_fields.append(
+                self._api.meshed_region_get_property_field_name(self, index)
+            )
+        return available_property_fields
+
+    def property_field(self, property_name):
+        """
+        Property field getter. It can be coordinates (field),
+        element types (property field)...
+
+        Returns
+        -------
+        field_or_property_field : core.Field or core.PropertyField
+        """
+        return self.field_of_properties(property_name)
+
+    @version_requires("3.0")
+    def set_property_field(self, property_name, value):
+        """
+        Property field setter. It can be coordinates (field),
+        element types (property field)...
+
+        Parameters
+        ----------
+        property_name : str
+            property name of the field to set
+        value : PropertyField or Field
+        """
+        if property_name is nodal_properties.coordinates:
+            self.set_coordinates_field(value)
+        else:
+            self._api.meshed_region_set_property_field(self, property_name, value)
+
+    @update_grid
+    @version_requires("3.0")
+    def set_coordinates_field(self, coordinates_field):
+        """
+        Coordinates field setter.
+
+        Parameters
+        ----------
+        coordinates_field : PropertyField or Field
+        """
+        self._api.meshed_region_set_coordinates_field(self, coordinates_field)
 
     @property
     def available_named_selections(self):
-        """List of available named selections.
+        """
+        List of available named selections.
 
         Returns
         -------
@@ -221,16 +314,24 @@ class MeshedRegion:
         return self._get_available_named_selections()
 
     def _get_available_named_selections(self):
-        """List of available named selections.
+        """
+        List of available named selections.
 
         Returns
         -------
         named_selections : list str
         """
-        return self._stub.List(self._message).named_selections
+        named_selections = []
+        n_selections = self._api.meshed_region_get_num_available_named_selection(self)
+        for index in range(n_selections):
+            named_selections.append(
+                self._api.meshed_region_get_named_selection_name(self, index)
+            )
+        return named_selections
 
     def named_selection(self, named_selection):
-        """Scoping containing the list of nodes or elements in the named selection.
+        """
+        Scoping containing the list of nodes or elements in the named selection.
 
         Parameters
         ----------
@@ -242,9 +343,9 @@ class MeshedRegion:
         named_selection : Scoping
         """
         if server_meet_version("2.1", self._server):
-            request = meshed_region_pb2.GetScopingRequest(mesh=self._message)
-            request.named_selection = named_selection
-            out = self._stub.GetScoping(request)
+            out = self._api.meshed_region_get_named_selection_scoping(
+                self, named_selection
+            )
             return scoping.Scoping(scoping=out, server=self._server)
         else:
             if hasattr(self, "_stream_provider"):
@@ -261,6 +362,21 @@ class MeshedRegion:
                     "model for server version 2.0. Please update your server."
                 )
 
+    @version_requires("3.0")
+    def set_named_selection_scoping(self, named_selection_name, scoping):
+        """
+        Named selection scoping setter.
+
+        Parameters
+        ----------
+        named_selection_name : str
+            named selection name
+        scoping : Scoping
+        """
+        return self._api.meshed_region_set_named_selection_scoping(
+            self, named_selection_name, scoping
+        )
+
     def _set_stream_provider(self, stream_provider):
         self._stream_provider = stream_provider
 
@@ -302,12 +418,47 @@ class MeshedRegion:
     #     name = None
     #     if self._name:
     #         name = 'Skin of %s' % self._name
-    #     self._message = skin.get_output(0, types.meshed_region)
+    #     self._internal_obj = skin.get_output(0, types.meshed_region)
     #     return MeshedRegion(self._server.channel, skin, self._model, name)
 
-    def _as_vtk(self, as_linear=True, include_ids=False):
+    def deform_by(self, deform_by, scale_factor=1.0):
+        """
+        Deforms the mesh according to a 3D vector field and an additional scale factor.
+
+        Parameters
+        ----------
+        deform_by : Field, FieldsContainer, Result, Operator
+            Used to deform the plotted mesh. Must output a unique 3D vector field.
+            Defaults to None.
+        scale_factor : float, Field, FieldsContainer, optional
+            Used to scale the mesh deformation. Defaults to 1.0. Can be a scalar Field
+            (or a FieldsContainer with only one Field) to get a spatially non-homogeneous scaling.
+        Returns
+        -------
+
+        """
+        from ansys.dpf.core.operators.math import add, scale, unit_convert
+        if hasattr(deform_by, "eval"):
+            # If a Result or an Operator, eval and get the field.
+            deform_by = deform_by.eval()[0]
+        if isinstance(deform_by, ansys.dpf.core.fields_container.FieldsContainer):
+            # If a FieldsContainer, take the field.
+            deform_by = deform_by[0]
+        if deform_by.unit != self.unit:
+            unit_convert(deform_by, self.unit)
+        scale_op = scale(field=deform_by, ponderation=scale_factor)
+        return add(
+            fieldA=self.nodes.coordinates_field, fieldB=scale_op.outputs.field
+        ).eval()
+
+    def _as_vtk(self, coordinates=None, as_linear=True, include_ids=False):
         """Convert DPF mesh to a PyVista unstructured grid."""
-        nodes = self.nodes.coordinates_field.data
+        if coordinates is None:
+            coordinates_field = self.nodes.coordinates_field
+            coordinates = self.nodes.coordinates_field.data
+        else:
+            coordinates_field = coordinates
+            coordinates = coordinates.data
         etypes = self.elements.element_types_field.data
         conn = self.elements.connectivities_field.data
         try:
@@ -318,18 +469,25 @@ class MeshedRegion:
                 "with :\n pip install pyvista>=0.24.0"
             )
 
-        grid = dpf_mesh_to_vtk(nodes, etypes, conn, as_linear)
+        grid = dpf_mesh_to_vtk(coordinates, etypes, conn, as_linear)
 
         # consider adding this when scoping request is faster
         if include_ids:
-            grid["node_ids"] = self.nodes.scoping.ids
-            grid["element_ids"] = self.elements.scoping.ids
+            self._nodeids = self.elements.scoping.ids
+            self._elementids = self.nodes.scoping.ids
+            grid["node_ids"] = self._elementids
+            grid["element_ids"] = self._nodeids
+
+        # Quick fix required to hold onto the data as PyVista does not make a copy.
+        # All of those now return DPFArrays
+        setattr(grid, "_dpf_cache", [coordinates, coordinates_field])
 
         return grid
 
     @property
     def grid(self):
-        """Unstructured grid in VTK fromat from PyVista.
+        """
+        Unstructured grid in VTK format from PyVista.
 
         Returns
         -------
@@ -340,7 +498,7 @@ class MeshedRegion:
         --------
         >>> import ansys.dpf.core as dpf
         >>> from ansys.dpf.core import examples
-        >>> model = dpf.Model(examples.static_rst)
+        >>> model = dpf.Model(examples.find_static_rst())
         >>> meshed_region = model.metadata.meshed_region
         >>> grid = meshed_region.grid
 
@@ -354,34 +512,31 @@ class MeshedRegion:
 
         """
         if self._full_grid is None:
-            self._full_grid = self._as_vtk()
+            self._full_grid = self._as_vtk(self.nodes.coordinates_field)
         return self._full_grid
 
     def plot(
-            self,
-            field_or_fields_container=None,
-            notebook=None,
-            shell_layers=None,
-            off_screen=None,
-            show_axes=True,
-            **kwargs
+        self,
+        field_or_fields_container=None,
+        shell_layers=None,
+        deform_by=None,
+        scale_factor=1.0,
+        **kwargs,
     ):
-        """Plot the field or fields container on the mesh.
+        """
+        Plot the field or fields container on the mesh.
 
         Parameters
         ----------
         field_or_fields_container : dpf.core.Field or dpf.core.FieldsContainer
             Field or fields container to plot. The default is ``None``.
-        notebook : bool, optional
-            Whether the plotting in the notebook is 2D or 3D. The default is
-            ``None``, in which case the plotting is 2D.
         shell_layers : core.shell_layers, optional
             Enum used to set the shell layers if the model to plot contains shell elements.
-        off_screen : bool, optional
-            Whether to render the plot off screen, which is useful for automated screenshots.
-            The default is "None", in which case the plot renders off screen.
-        show_axes : bool, optional
-            Whether to show a VTK axes widget. The default is ``True``.
+        deform_by : Field, Result, Operator, optional
+            Used to deform the plotted mesh. Must output a 3D vector field.
+            Defaults to None.
+        scale_factor : float, optional
+            Scaling factor to apply when warping the mesh. Defaults to 1.0.
         **kwargs : optional
             Additional keyword arguments for the plotter. For additional keyword
             arguments, see ``help(pyvista.plot)``.
@@ -392,29 +547,37 @@ class MeshedRegion:
 
         >>> import ansys.dpf.core as dpf
         >>> from ansys.dpf.core import examples
-        >>> model = dpf.Model(examples.static_rst)
+        >>> model = dpf.Model(examples.find_static_rst())
         >>> disp = model.results.displacement()
         >>> field = disp.outputs.fields_container()[0]
         >>> model.metadata.meshed_region.plot(field)
 
         """
-        pl = _DpfPlotter(self)
         if field_or_fields_container is not None:
+            pl = Plotter(self, **kwargs)
             return pl.plot_contour(
                 field_or_fields_container,
-                notebook,
                 shell_layers,
-                off_screen,
-                show_axes,
-                **kwargs
+                show_axes=kwargs.pop("show_axes", True),
+                deform_by=deform_by,
+                scale_factor=scale_factor,
+                **kwargs,
             )
 
-        # otherwise, simply plot self
-        kwargs["notebook"] = notebook
-        return pl.plot_mesh(**kwargs)
+        # otherwise, simply plot the mesh
+        pl = DpfPlotter(**kwargs)
+        pl.add_mesh(
+            self,
+            deform_by=deform_by,
+            scale_factor=scale_factor,
+            show_axes=kwargs.pop("show_axes", True),
+            **kwargs,
+        )
+        return pl.show_figure(**kwargs)
 
     def deep_copy(self, server=None):
-        """Create a deep copy of the meshed region's data on a given server.
+        """
+        Create a deep copy of the meshed region's data on a given server.
 
         This method is useful for passing data from one server instance to another.
 
@@ -438,12 +601,14 @@ class MeshedRegion:
         --------
         >>> import ansys.dpf.core as dpf
         >>> from ansys.dpf.core import examples
-        >>> model = dpf.Model(examples.static_rst)
+        >>> model = dpf.Model(examples.find_static_rst())
         >>> meshed_region = model.metadata.meshed_region
         >>> other_server = dpf.start_local_server(as_global=False)
         >>> deep_copy = meshed_region.deep_copy(server=other_server)
 
         """
+        if self.nodes.scoping is None: # empty Mesh
+            return MeshedRegion()
         node_ids = self.nodes.scoping.ids
         element_ids = self.elements.scoping.ids
         mesh = MeshedRegion(
@@ -462,16 +627,46 @@ class MeshedRegion:
         mesh.unit = self.unit
         return mesh
 
-    def __send_init_request(self, num_nodes=0, num_elements=0):
-        request = meshed_region_pb2.CreateRequest()
-        if num_nodes:
-            request.num_nodes_reserved = num_nodes
-        if num_elements:
-            request.num_elements_reserved = num_elements
-        self._message = self._stub.Create(request)
+    def field_of_properties(self, property_name):
+        """
+        Returns the ``Field`` or ``PropertyField`` associated
+        to a given property of the mesh
 
-    _to_cache = {
-        _get_unit: [_set_unit],
-        _get_available_named_selections: None,
-        named_selection: None
-    }
+        Parameters
+        ----------
+        property_name : str, common.elemental_properties, common.nodal_properties
+            Name of the property.
+
+        Returns
+        -------
+        properties : Field, PropertyField
+
+        Examples
+        --------
+        >>> import ansys.dpf.core as dpf
+        >>> from ansys.dpf.core import examples
+        >>> model = dpf.Model(examples.find_static_rst())
+        >>> meshed_region = model.metadata.meshed_region
+        >>> connectivity = meshed_region.field_of_properties(
+        ...     dpf.common.elemental_properties.connectivity)
+        >>> coordinates = meshed_region.field_of_properties(dpf.common.nodal_properties.coordinates)
+        """
+        if property_name is nodal_properties.coordinates:
+            field_out = self._api.meshed_region_get_coordinates_field(self)
+            return field.Field(server=self._server, field=field_out)
+        else:
+            field_out = self._api.meshed_region_get_property_field(self, property_name)
+            if isinstance(field_out, int):
+                res = property_field.PropertyField(
+                    server=self._server, property_field=field_out
+                )
+                return res
+            else:
+                if field_out.datatype == "int":
+                    return property_field.PropertyField(
+                        server=self._server, property_field=field_out
+                    )
+                else:
+                    # Not sure we go through here since the only datatype not int is coordinates,
+                    # which is already dealt with previously.
+                    return field.Field(server=self._server, field=field_out)
