@@ -1,22 +1,14 @@
-"""
-.. _ref_operators_package:
-
-ansys.dpf.core.operators
-========================
-"""
-  
-
-from . import result
 from . import math
+from . import result
+from . import utility
 from . import min_max
 from . import scoping
-from . import utility
-from . import metadata
-from . import logic
-from . import mesh
 from . import filter
+from . import logic
+from . import metadata
 from . import serialization
-from . import averaging
+from . import mesh
 from . import geo
+from . import averaging
 from . import invariant
 from . import mapping
