@@ -9,8 +9,5 @@ from .gauss_to_node import gauss_to_node
 from .elements_facets_surfaces_over_time import elements_facets_surfaces_over_time 
 from .elements_volume import elements_volume 
 from .element_nodal_contribution import element_nodal_contribution 
-from .center_of_gravity import center_of_gravity 
 from .integrate_over_elements import integrate_over_elements 
-from .mass import mass 
-from .moment_of_inertia import moment_of_inertia 
 from .normals import normals 
