@@ -1,7 +1,9 @@
 from .eigen_values import eigen_values 
 from .principal_invariants import principal_invariants 
 from .von_mises_eqv import von_mises_eqv 
+from .segalman_von_mises_eqv import segalman_von_mises_eqv 
 from .von_mises_eqv_fc import von_mises_eqv_fc 
+from .segalman_von_mises_eqv_fc import segalman_von_mises_eqv_fc 
 from .invariants import invariants 
 from .eigen_values_fc import eigen_values_fc 
 from .invariants_fc import invariants_fc 
