@@ -4,9 +4,10 @@ from .connectivity_ids import connectivity_ids
 from .on_property import on_property 
 from .transpose import transpose 
 from .rescope_fc import rescope_fc 
-from .splitted_on_property_type import splitted_on_property_type 
 from .rescope import rescope 
+from .split_on_property_type import split_on_property_type 
 from .on_named_selection import on_named_selection 
 from .nodal_from_mesh import nodal_from_mesh 
 from .change_fc import change_fc 
+from .on_mesh_property import on_mesh_property 
 from .from_mesh import from_mesh 
