@@ -1,30 +1,28 @@
 """
+.. _ref_timefreqsupport:
+
 TimeFreqSupport
 ===============
 """
+from ansys.dpf.gate import time_freq_support_capi, time_freq_support_grpcapi
+
 from ansys import dpf
 from ansys.dpf import core
 from ansys.dpf.core import errors as dpf_errors
-from ansys.dpf.core.errors import protect_grpc
-from ansys.grpc.dpf import (
-    base_pb2,
-    support_pb2,
-    time_freq_support_pb2,
-    time_freq_support_pb2_grpc,
-)
+from ansys.dpf.core.support import Support
 
 
-class TimeFreqSupport:
-    """Represents a time frequency support, which is a description of a temporal or frequency analysis.
+class TimeFreqSupport(Support):
+    """Represents a time frequency support, a description of a temporal or frequency analysis.
 
     This class stores values such as the frequencies (time/complex), RPMs, and harmonic indices.
     The RPM value is a step (or load step)-based value.
-    The time freqencies, complex frequencies, and harmonic indices are set-based values.
+    The time frequencies, complex frequencies, and harmonic indices are set-based values.
     There is one set value for each step/substep combination.
 
     Parameters
     ----------
-    time_freq_support : ansys.grpc.dpf.time_freq_support_pb2.TimeFreqSupport
+    time_freq_support : ctypes.c_void_p, ansys.grpc.dpf.time_freq_support_pb2.TimeFreqSupport
     server : ansys.dpf.core.server, optional
         Server with the channel connected to the remote or local instance.
         The default is ``None``, in which case an attempt is made to use the
@@ -44,19 +42,27 @@ class TimeFreqSupport:
 
     def __init__(self, time_freq_support=None, server=None):
         """Initialize the TimeFreqSupport with its TimeFreqSupport message (if possible)."""
-        if server is None:
-            server = dpf.core._global_server()
+        super(TimeFreqSupport, self).__init__(support=time_freq_support, server=server)
 
-        self._server = server
-        self._stub = self._connect()
-        if isinstance(time_freq_support, time_freq_support_pb2.TimeFreqSupport):
-            self._message = time_freq_support
-        elif isinstance(time_freq_support, support_pb2.Support):
-            self._message = time_freq_support_pb2.TimeFreqSupport()
-            self._message.id = time_freq_support.id
+        # step 2: get api
+        self._api = self._server.get_api_for_type(
+            capi=time_freq_support_capi.TimeFreqSupportCAPI,
+            grpcapi=time_freq_support_grpcapi.TimeFreqSupportGRPCAPI)
+
+        # step3: init environment
+        self._api.init_time_freq_support_environment(self)  # creates stub when gRPC
+
+        # step4: if object exists: take instance, else create it:
+        # object_name -> protobuf.message, DPFObject*
+        if time_freq_support is not None:
+            self._internal_obj = time_freq_support
+            # Might to test for type for CLayer as I have not tested this for C
+            #self._internal_obj = support_api.support_get_as_time_freq_support(self)
         else:
-            request = base_pb2.Empty()
-            self._message = self._stub.Create(request)
+            if self._server.has_client():
+                self._internal_obj = self._api.time_freq_support_new_on_client(self._server.client)
+            else:
+                self._internal_obj = self._api.time_freq_support_new()
 
     def __str__(self):
         """Describe the entity.
@@ -67,7 +73,7 @@ class TimeFreqSupport:
         """
         from ansys.dpf.core.core import _description
 
-        return _description(self._message, self._server)
+        return _description(self._internal_obj, self._server)
 
     @property
     def time_frequencies(self):
@@ -83,7 +89,8 @@ class TimeFreqSupport:
         >>> time_freq_support = model.metadata.time_freq_support
         >>> freq = time_freq_support.time_frequencies
         >>> freq.data
-        array([0.        , 0.019975  , 0.039975  , 0.059975  , 0.079975  ,
+        <BLANKLINE>
+        ...rray([0.        , 0.019975  , 0.039975  , 0.059975  , 0.079975  ,
                0.099975  , 0.119975  , 0.139975  , 0.159975  , 0.179975  ,
                0.199975  , 0.218975  , 0.238975  , 0.258975  , 0.278975  ,
                0.298975  , 0.318975  , 0.338975  , 0.358975  , 0.378975  ,
@@ -103,10 +110,7 @@ class TimeFreqSupport:
         frequencies: Field
             Field of time frequencies that must be set.
         """
-        request = time_freq_support_pb2.TimeFreqSupportUpdateRequest()
-        request.time_freq_support.CopyFrom(self._message)
-        request.freq_real.CopyFrom(frequencies._message)
-        self._stub.Update(request)
+        self._api.time_freq_support_set_shared_time_freqs(self, frequencies)
 
     @time_frequencies.setter
     def time_frequencies(self, value):
@@ -118,7 +122,7 @@ class TimeFreqSupport:
         value : Field
             Field of time frequencies that must be set.
         """
-        return self._set_time_frequencies(value)
+        self._set_time_frequencies(value)
 
     @property
     def complex_frequencies(self):
@@ -146,10 +150,7 @@ class TimeFreqSupport:
         complex_frequencies : Field
             Field of frequencies that must be set.
         """
-        request = time_freq_support_pb2.TimeFreqSupportUpdateRequest()
-        request.time_freq_support.CopyFrom(self._message)
-        request.freq_complex.CopyFrom(complex_frequencies._message)
-        self._stub.Update(request)
+        self._api.time_freq_support_set_shared_imaginary_freqs(self, complex_frequencies)
 
     @complex_frequencies.setter
     def complex_frequencies(self, value):
@@ -161,7 +162,7 @@ class TimeFreqSupport:
         value : Field
             Field of complex frequencies that must be set.
         """
-        return self._set_complex_frequencies(value)
+        self._set_complex_frequencies(value)
 
     @property
     def rpms(self):
@@ -181,10 +182,7 @@ class TimeFreqSupport:
         rpms : Field
             Field of RPMs that must be set.
         """
-        request = time_freq_support_pb2.TimeFreqSupportUpdateRequest()
-        request.time_freq_support.CopyFrom(self._message)
-        request.rpm.CopyFrom(rpms._message)
-        self._stub.Update(request)
+        self._api.time_freq_support_set_shared_rpms(self, rpms)
 
     @rpms.setter
     def rpms(self, value):
@@ -224,13 +222,7 @@ class TimeFreqSupport:
         stage_num: int, default: 0, optional
             Stage number that is defined by these harmonic indices.
         """
-        request = time_freq_support_pb2.TimeFreqSupportUpdateRequest()
-        cyclic_data = time_freq_support_pb2.CyclicHarmonicData()
-        request.time_freq_support.CopyFrom(self._message)
-        cyclic_data.cyc_harmonic_index.CopyFrom(harmonic_indices._message)
-        cyclic_data.stage_num = stage_num
-        request.cyc_harmonic_data.CopyFrom(cyclic_data)
-        self._stub.Update(request)
+        self._api.time_freq_support_set_harmonic_indices(self, harmonic_indices, stage_num)
 
     @property
     def n_sets(self):
@@ -271,21 +263,28 @@ class TimeFreqSupport:
         """
         return self._get_frequency(step, substep, cumulative_index, cplx)
 
-    @protect_grpc
     def _get_frequency(self, step, substep, cumulative_index, cplx):
         """Retrieves the frequence corresponding to the requested step/substep or
         cumulative index.
         """
-        request = time_freq_support_pb2.GetRequest()
-        request.time_freq_support.CopyFrom(self._message)
-        request.complex = cplx
-        # request.complex = base_pb2.Complex.Value(cplx)
         if cumulative_index is None:
-            request.step_substep.step = step
-            request.step_substep.substep = substep
+            # Use by_step methods
+            if cplx:
+                # Call for imaginary
+                return self._api.time_freq_support_get_imaginary_freq_by_step(self, step, substep)
+            else:
+                # Call for real
+                return self._api.time_freq_support_get_time_freq_by_step(self, step, substep)
         else:
-            request.cumulative_index = cumulative_index
-        return self._stub.Get(request).frequency
+            # Use by_cumul_index methods
+            if cplx:
+                # Call for imaginary
+                return self._api.time_freq_support_get_imaginary_freq_by_cumul_index(
+                    self, cumulative_index)
+            else:
+                # Call for real
+                return self._api.time_freq_support_get_time_freq_by_cumul_index(self,
+                                                                                cumulative_index)
 
     def get_cumulative_index(self, step=0, substep=0, freq=None, cplx=False):
         """Retrieves the cumulative index corresponding to the requested step/substep
@@ -300,7 +299,7 @@ class TimeFreqSupport:
         freq  : double, optional
             Frequency in Hz.
         cplx : False, optional
-            Whehter to return a complex frequency. The default is ``False``.
+            Whether to return a complex frequency. The default is ``False``.
 
         Returns
         -------
@@ -310,34 +309,37 @@ class TimeFreqSupport:
         """
         return self._get_cumulative_index(step, substep, freq, cplx)
 
-    @protect_grpc
     def _get_cumulative_index(self, step, substep, freq, cplx):
         """Retrieve the cumulative index corresponding to the requested step/substep
         or frequency."""
-        request = time_freq_support_pb2.GetRequest()
-        request.time_freq_support.CopyFrom(self._message)
-        request.bool_cumulative_index = True
-        request.complex = cplx
-        if freq is not None:
-            request.frequency = freq
+        if freq is None:
+            if cplx is False:
+                return self._api.time_freq_support_get_time_freq_cummulative_index_by_step(self,
+                                                                                           step,
+                                                                                           substep)
+            else:
+                raise NotImplementedError("get_cumulative_index is not implemented for cplx=False")
         else:
-            request.step_substep.step = step
-            request.step_substep.substep = substep
-        return self._stub.Get(request).cumulative_index
+            from ansys.dpf.gate import integral_types
+            i1 = integral_types.MutableInt32()
+            i2 = integral_types.MutableInt32()
+            if cplx:
+                return self._api.time_freq_support_get_imaginary_freqs_cummulative_index(
+                    self, freq, i1, i2
+                )
+            else:
+                return self._api.time_freq_support_get_time_freq_cummulative_index_by_value(
+                    self, freq, i1, i2
+                )
 
-    @protect_grpc
     def _sets_count(self):
         """
         Returns
         -------
         count : int
         """
-        request = time_freq_support_pb2.CountRequest()
-        request.time_freq_support.CopyFrom(self._message)
-        request.entity = base_pb2.NUM_SETS
-        return self._stub.Count(request).count
+        return self._api.time_freq_support_get_number_sets(self)
 
-    @protect_grpc
     def _get_frequencies(self, cplx=False):
         """Retrieves a field of all the frequencies in the model
         (complex or real).
@@ -352,17 +354,18 @@ class TimeFreqSupport:
         field : dpf.core.Field
             Field of all the frequencies in the model (complex or real).
         """
-        request = time_freq_support_pb2.ListRequest()
-        request.time_freq_support.CopyFrom(self._message)
 
-        list_response = self._stub.List(request)
-        if cplx is True and list_response.freq_complex.id != 0:
-            return dpf.core.Field(server=self._server, field=list_response.freq_complex)
-        elif cplx is False and list_response.freq_real.id != 0:
-            return dpf.core.Field(server=self._server, field=list_response.freq_real)
-        return None
+        # attributes_list = self._get_attributes_list()
+        if cplx:  # and "freq_complex" in attributes_list:
+            # return attributes_list["freq_complex"]
+            freq = self._api.time_freq_support_get_shared_imaginary_freqs(self)
+        # elif cplx != True and "freq_real" in attributes_list:
+        else:
+            # return attributes_list["freq_real"]
+            freq = self._api.time_freq_support_get_shared_time_freqs(self)
+        if freq is not None:
+            return dpf.core.Field(server=self._server, field=freq)
 
-    @protect_grpc
     def _get_rpms(self):
         """Retrieves a field of all the RPMs in the model.
 
@@ -371,15 +374,10 @@ class TimeFreqSupport:
         field : dpf.core.Field
             Field of all the RPMs in the model (complex or real).
         """
-        request = time_freq_support_pb2.ListRequest()
-        request.time_freq_support.CopyFrom(self._message)
+        rpm = self._api.time_freq_support_get_shared_rpms(self)
+        if rpm is not None:
+            return dpf.core.Field(server=self._server, field=rpm)
 
-        list_response = self._stub.List(request)
-        if list_response.rpm.id != 0:
-            return dpf.core.Field(server=self._server, field=list_response.rpm)
-        return None
-
-    @protect_grpc
     def _get_harmonic_indices(self, stage_num=0):
         """Retrieves a field of all the harmonic indices in the model.
 
@@ -391,28 +389,21 @@ class TimeFreqSupport:
         stage_num: int, optional, default = 0
             Targeted stage number.
         """
-        request = time_freq_support_pb2.ListRequest()
-        request.time_freq_support.CopyFrom(self._message)
-        request.cyclic_stage_num = stage_num
-
-        list_response = self._stub.List(request)
-        if list_response.cyc_harmonic_index.id != 0:
-            return dpf.core.Field(
-                server=self._server, field=list_response.cyc_harmonic_index
-            )
-        return None
+        harmonic_indices = self._api.time_freq_support_get_shared_harmonic_indices(self, stage_num)
+        if harmonic_indices is not None:
+            return dpf.core.Field(server=self._server, field=harmonic_indices)
 
     def append_step(
-        self,
-        step_id,
-        step_time_frequencies,
-        step_complex_frequencies=None,
-        rpm_value=None,
-        step_harmonic_indices=None,
+            self,
+            step_id,
+            step_time_frequencies,
+            step_complex_frequencies=None,
+            rpm_value=None,
+            step_harmonic_indices=None,
     ):
         """Append a step with all its field values in the time frequencies support.
         The RPM value is a step (or load step)-based value.
-        The values for time freqencies, complex frequencies, and harmonic indices are set-based.
+        The values for time frequencies, complex frequencies, and harmonic indices are set-based.
         There is one set value for each step/substep combination.
 
         It is necessary that each call of my_time_freq_support.append_step(kwargs**) contains
@@ -502,7 +493,7 @@ class TimeFreqSupport:
     def deep_copy(self, server=None):
         """Create a deep copy of the data for a time frequency support on a given server.
 
-        This methos is useful for passing data from one server instance to another.
+        This method is useful for passing data from one server instance to another.
 
         Parameters
         ----------
@@ -552,15 +543,3 @@ class TimeFreqSupport:
             harmonic_indices.scoping.location = core.locations.time_freq_step
         harmonic_indices.append(step_harmonic_indices, step_id)
         self.set_harmonic_indices(harmonic_indices, stage_num)
-
-    def _connect(self):
-        """Connect to the gRPC service."""
-        return time_freq_support_pb2_grpc.TimeFreqSupportServiceStub(
-            self._server.channel
-        )
-
-    def __del__(self):
-        try:
-            self._stub.Delete(self._message)
-        except:
-            pass
